"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const object_graph_1 = require("./object-graph");
const binary_type_identifier_1 = require("./binary-type-identifier");
const circularReferenceScalar = new object_graph_1.CircularReferenceScalar();
const emptyArrayScalar = new object_graph_1.EmptyArrayScalar();
const notScalar = {};
class YamprintGraphBuilder {
    constructor(rules) {
        this.rules = rules;
        this._knownNodes = new Set();
        this._currentDepth = -1;
    }
    /**
     * Takes an object and tries to identify it as a scalar. Returns a `Node` (could be a literal) if it was successful, or `null` if not.
     *
     * If you want to add more scalar types, you should do it here.
     * @param value
     * @returns {Node}
     * @private
     */
    _tryBuildScalar(value) {
        if (value == null)
            return value;
        if ([String, Number, Boolean, Symbol, Function].some(ctor => value instanceof ctor)) {
            value = value.valueOf();
        }
        if (typeof value === "string") {
            let lines = value.split(/(?:\r\n|\r|\n)/);
            if (lines.length === 1)
                return value;
            return new object_graph_1.TextBlockScalar(lines);
        }
        if (typeof value !== "object")
            return value;
        if ([Date, RegExp].find(ctor => value instanceof ctor)) {
            return value;
        }
        let bt = binary_type_identifier_1.BinaryTypeIdentifier.getBinaryTypeInfo(value);
        if (bt)
            return new object_graph_1.BinaryScalar(bt.name, bt.size);
        return notScalar;
    }
    _readObjectProprties(instance) {
        if (this.rules.maxDepth <= this._currentDepth) {
            return new object_graph_1.EmptyObjectScalar(this.rules.getConstructor(instance)).withMetadata({
                depthExceeded: true
            });
        }
        let resolveKey = (key) => {
            try {
                var value = instance[key];
            }
            catch (err) {
                return new object_graph_1.ThrewErrorScalar(err);
            }
            return this._toGraph(value);
        };
        let allProperties = new Map();
        let reachedMaxProperties = false;
        let registerAllKeys = (definingObject, objectDepth) => {
            let names = Object.getOwnPropertyNames(definingObject);
            for (let name of names) {
                //when there is an overriden property on the prototype
                if (allProperties.has(name))
                    continue;
                let propInfo = {
                    name: name,
                    descriptor: Object.getOwnPropertyDescriptor(definingObject, name),
                    owner: definingObject,
                    treeDepth: this._currentDepth,
                    objectDepth: objectDepth
                };
                if (!this.rules.propertyFilter(propInfo))
                    continue;
                if (allProperties.size >= this.rules.maxObjectLength) {
                    reachedMaxProperties = true;
                    return;
                }
                allProperties.set(name, propInfo);
            }
            let proto = Object.getPrototypeOf(definingObject);
            if (proto && this.rules.isPrototypeExplorable(proto)) {
                registerAllKeys(proto, objectDepth + 1);
            }
        };
        registerAllKeys(instance, 0);
        let properties = [];
        for (let [name, property] of allProperties) {
            if (property.descriptor.get && !this.rules.resolveGetters) {
                properties.push({
                    name: name,
                    value: new object_graph_1.UnresolvedGetterScalar()
                });
            }
            else {
                properties.push({
                    name: name,
                    value: resolveKey(name)
                });
            }
        }
        if (properties.length === 0) {
            return new object_graph_1.EmptyObjectScalar(this.rules.getConstructor(instance));
        }
        return new object_graph_1.ObjectNode(this.rules.getConstructor(instance), properties).withMetadata({
            sizeExceeded: reachedMaxProperties
        });
    }
    _readArray(value) {
        if (this.rules.maxDepth <= this._currentDepth) {
            return new object_graph_1.EmptyArrayScalar().withMetadata({
                depthExceeded: true
            });
        }
        let lastIndex = -1;
        let isSparse = value.some((x, index) => {
            if (index !== lastIndex + 1)
                return true;
            lastIndex++;
            return false;
        });
        if (isSparse) {
            let keys = Object.keys(value);
            let maxSizeReached = false;
            let propertyItems = [];
            for (let key of keys) {
                if (isNaN(+key))
                    continue;
                if (propertyItems.length >= this.rules.maxObjectLength) {
                    maxSizeReached = true;
                    break;
                }
                let innerValue = this._toGraph(value[key]);
                propertyItems.push(new object_graph_1.PropertyItem(key, innerValue));
            }
            return new object_graph_1.SparseArrayNode(propertyItems).withMetadata({
                sizeExceeded: maxSizeReached
            });
        }
        else if (lastIndex === -1) {
            return emptyArrayScalar;
        }
        else {
            let maxSizeReached = false;
            let graphItems = [];
            for (let item of value) {
                if (graphItems.length >= this.rules.maxObjectLength) {
                    maxSizeReached = true;
                    break;
                }
                graphItems.push(this._toGraph(item));
            }
            return new object_graph_1.ArrayNode(graphItems).withMetadata({
                sizeExceeded: maxSizeReached
            });
        }
    }
    _toGraph(instance) {
        this._currentDepth++;
        let result;
        let scalar = this._tryBuildScalar(instance);
        if (scalar !== notScalar) {
            result = scalar;
        }
        else if (this._knownNodes.has(instance)) {
            result = circularReferenceScalar;
        }
        else {
            this._knownNodes.add(instance);
            if (Array.isArray(instance)) {
                result = this._readArray(instance);
            }
            else {
                result = this._readObjectProprties(instance);
            }
            this._knownNodes.delete(instance);
        }
        this._currentDepth--;
        return result;
    }
    toGraph(instance) {
        try {
            return this._toGraph(instance);
        }
        finally {
            //even if GraphBuilder errors, we need to restore its neutral state so it doesn't get totally broken.
            this._currentDepth = -1;
            this._knownNodes.clear();
        }
    }
}
exports.YamprintGraphBuilder = YamprintGraphBuilder;
//# sourceMappingURL=graph-builder.js.map