"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const object_graph_1 = require("./object-graph");
const indented_writer_1 = require("./indented-writer");
class YamprintGraphPrinter {
    constructor(formatter) {
        this.formatter = formatter;
        this._writer = new indented_writer_1.IndentedWriter(formatter.indent);
    }
    _formatLiteral(literal) {
        if (literal === null)
            return this.formatter.null;
        if (literal === undefined)
            return this.formatter.undefined;
        switch (typeof literal) {
            case "boolean":
                return this.formatter.boolean(literal);
            case "string":
                return this.formatter.string(literal);
            case "symbol":
                return this.formatter.symbol(literal);
            case "number":
                return this.formatter.number(literal);
            case "function":
                return this.formatter.function(literal);
        }
        if (literal instanceof Date) {
            return this.formatter.date(literal);
        }
        else if (literal instanceof RegExp) {
            return this.formatter.regexp(literal);
        }
        else {
            return null;
        }
    }
    _printArray(instance) {
        if (instance.metadata.depthExceeded) {
            this._writer.writeLine(this.formatter.arrayDepthExceeded);
            return;
        }
        let lines = [];
        for (let item of instance.items) {
            let result = this._formatScalar(item);
            this._writer.write(this.formatter.arrayPrefix);
            if (result != null) {
                this._writer.writeLine(result);
            }
            else {
                this._writer.indent(1);
                this._printComplex(item);
                this._writer.indent(-1);
            }
        }
        if (instance.metadata.sizeExceeded) {
            this._writer.writeLine(this.formatter.sizeExceededToken);
        }
    }
    _printSparseArray(node) {
        let lines = [];
        for (let item of node.items) {
            let result = this._formatScalar(item.value);
            this._writer.write(`(${item.name}) ${this.formatter.arrayPrefix}`);
            if (result != null) {
                this._writer.writeLine(result);
            }
            else {
                this._writer.indent(1);
                this._printComplex(item.value);
                this._writer.indent(-1);
            }
        }
        if (node.metadata.sizeExceeded) {
            this._writer.writeLine(this.formatter.sizeExceededToken);
        }
    }
    _formatScalar(scalar) {
        if (scalar instanceof object_graph_1.NodeBase) {
            if (scalar instanceof object_graph_1.BinaryScalar) {
                return this.formatter.binary(scalar);
            }
            else if (scalar instanceof object_graph_1.CircularReferenceScalar) {
                return this.formatter.circularReference;
            }
            else if (scalar instanceof object_graph_1.EmptyArrayScalar) {
                return this.formatter.emptyArray;
            }
            else if (scalar instanceof object_graph_1.EmptyObjectScalar) {
                return this.formatter.emptyObject(scalar.ctor, scalar.metadata);
            }
            else if (scalar instanceof object_graph_1.UnresolvedGetterScalar) {
                return this.formatter.unreoslvedGetter;
            }
            else if (scalar instanceof object_graph_1.ThrewErrorScalar) {
                return this.formatter.threwAlert(scalar.error);
            }
        }
        else {
            return this._formatLiteral(scalar);
        }
    }
    _printTextBlock(block) {
        return this.formatter.textBlock(block);
    }
    _printObject(node) {
        let tag = this.formatter.constructorTag(node.ctor);
        if (tag)
            this._writer.writeLine(tag);
        for (let property of node.properties) {
            let tryScalar = this._formatScalar(property.value);
            let name = this.formatter.propertyKey(property.name);
            if (tryScalar != null) {
                this._writer.writeLine(`${name}${tryScalar}`);
            }
            else {
                this._writer.writeLine(`${name}`);
                this._writer.indent(1);
                this._printComplex(property.value);
                this._writer.indent(-1);
            }
        }
        if (node.metadata.sizeExceeded) {
            this._writer.writeLine(this.formatter.sizeExceededToken);
        }
    }
    _printComplex(node) {
        if (node instanceof object_graph_1.ObjectNode) {
            this._printObject(node);
        }
        else if (node instanceof object_graph_1.ArrayNode) {
            this._printArray(node);
        }
        else if (node instanceof object_graph_1.SparseArrayNode) {
            this._printSparseArray(node);
        }
        else if (node instanceof object_graph_1.TextBlockScalar) {
            this._printTextBlock(node);
        }
    }
    print(node) {
        try {
            let scalar = this._formatScalar(node);
            if (scalar != null) {
                this._writer.writeLine(scalar);
            }
            else {
                this._printComplex(node);
            }
            let result = this._writer.output();
            return result;
        }
        finally {
            this._writer.clear();
        }
    }
}
exports.YamprintGraphPrinter = YamprintGraphPrinter;
//# sourceMappingURL=graph-printer.js.map